"""Handle point-and-click element selection in a Reflex app"""

import functools
import inspect
from pathlib import Path
from typing import Any

import reflex as rx

# These components will never be selectable.
ignored_components = [
    rx.accordion.root.__self__,
    rx.accordion.icon.__self__,
    rx.accordion.item.__self__,
    rx.accordion.trigger.__self__,
    rx.alert_dialog.trigger.__self__,
    rx.hover_card.root.__self__,
    rx.hover_card.trigger.__self__,
    rx.hover_card.content.__self__,
    rx.popover.root.__self__,
    rx.popover.trigger.__self__,
    rx.popover.content.__self__,
    rx.recharts.responsive_container.__self__,
    rx.select.content.__self__,
    rx.select.group.__self__,
    rx.select.item.__self__,
    rx.select.root.__self__,
    rx.theme.__self__,
    rx.Fragment,
]
# Radix triggers that fire on_pointer_down events need to be disabled so that on_click works.
neuter_pointer_down = [
    rx.select.trigger.__self__,
]
# For components that do not support on_click at all, wrap these in a box.
box_wrap_components = [
    rx.markdown.__self__,
]
# Preserve a reference to the original create method's function.
og_component_create = rx.Component.create.__func__


class ClickSelectionState(rx.State):
    """Track the currently selected component and the code that created it."""

    selected_element: str = ""
    _selected_code: list[str] = []
    filename: str = ""
    selected_function : str = ""
    start_lineno: int
    end_lineno: int
    selected_module: str = ""

    @rx.var
    def code(self) -> str:
        return "\n".join(self._selected_code)

    async def handle_click(self, filename, function_name, lineno, end_lineno):
        self.selected_element = f"{filename}:{lineno}-{end_lineno}"
        self.filename = filename
        self.selected_function = function_name
        self.start_lineno = lineno
        self.end_lineno = end_lineno
        import importlib
        module = importlib.import_module(filename)
        self.selected_module = module.__name__
        self._selected_code = (
            Path(module.__file__).read_text().splitlines()[lineno - 1 : end_lineno]
        )


def _on_click_handler(
    props: dict[str, Any], filename: str, function_name: str, lineno: int, end_lineno: int
):
    """Generate on_click handler and outline props for selectable components.

    Args:
        filename: The filename of the component.
        function_name: The function name of the component.
        lineno: The starting line number of the component.
        end_lineno: The ending line number of the component

    Returns:
        A dictionary of props that includes the on_click handler and outline props.
    """
    return {
        "on_click": ClickSelectionState.handle_click(
            filename,
            function_name,
            lineno,
            end_lineno,
        ).prevent_default.stop_propagation,
        "outline": rx.cond(
            ClickSelectionState.selected_element == f"{filename}:{lineno}-{end_lineno}",
            "2px dotted var(--ruby-7)",
            props.get("outline", "none"),
        ),
    }


def component_create_override(base_paths: list[Path]) -> classmethod:
    """Generate an override for Component.create that is active for the given paths.

    Args:
        base_paths: Components originating from modules prefixed by these paths will be clickable.

    Returns:
        A replacement Component.create function that adds an on_click handler where needed.
    """

    @classmethod
    def _component_create_override(cls, *children, **props):
        if cls not in ignored_components:
            # Walk up the stack to find the first frame that originates from a base path.
            for frame in inspect.stack()[1:]:
                if any(frame.filename.startswith(str(p)) for p in base_paths):
                    module = inspect.getmodule(frame.frame).__name__
                    if cls in box_wrap_components:
                        return rx.box(
                            og_component_create(cls, *children, **props),
                            **_on_click_handler(
                                props,
                                # frame.filename,
                                module,
                                frame.function,
                                frame.lineno,
                                frame.positions.end_lineno,
                            ),
                        )
                    props.update(
                        _on_click_handler(
                            props,
                            # frame.filename,
                            module,
                            frame.function,
                            frame.lineno,
                            frame.positions.end_lineno,
                        )
                    )
                    if cls in neuter_pointer_down:
                        props.setdefault("special_props", set()).add(
                            rx.Var.create(
                                "onPointerDown={(e) => e.preventDefault()}",
                                _var_is_string=False,
                            ),
                        )
                    break
        return og_component_create(cls, *children, **props)

    return _component_create_override


def clickable(base_paths: list[Path] | None = None):
    """A decorator helper to make all components in a given page clickable to select.

    The active selection (filename and line range) is stored in
    ClickSelectionState.selected_element

    The code for the selection is cached in ClickSelectionState._selected_code

    Args:
        base_paths: Components originating from modules prefixed by these paths will be clickable.

    Returns:
        A decorator that adds an on_click handler to the component.
    """
    if base_paths is None:
        base_paths = [Path(".").resolve()]

    def outer(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            # Override the component create method.
            rx.Component.create = component_create_override(base_paths=base_paths)
            try:
                # Call the page wrapper
                return fn(*args, **kwargs)
            finally:
                # Restore the original component create method.
                rx.Component.create = classmethod(og_component_create)

        # Preserve the original function name to avoid having to specify route in add_page.
        inner.__name__ = fn.__name__
        return inner

    return outer



"""The agent that runs on the user's local client."""

import ast
import importlib
import inspect

import black
from flexai import Agent
from flexai.message import Message, ToolResultMessage
from pydantic import BaseModel


# NOTE: using BaseModel here instead of rx.Base due to FastAPI not liking the v1 models
class InternRequest(BaseModel):
    """The request to the AI intern."""

    prompt: str
    selected_code: str
    selected_module: str
    selected_function: str

class InternResponse(BaseModel):
    """The response from the AI intern."""

    request_id: str
    messages: list[Message]

class ToolRequestResponse(BaseModel):
    """The response from the tool to the AI intern."""

    request_id: str
    messages: list[ToolResultMessage]



def get_agent() -> Agent:
    """Get an instance of an intern."""
    return Agent(
        tools=[
            get_function_source,
            create_function,
            update_function,
        ],
    )

def get_module_ast(module_name: str) -> ast.Module:
    """Get the AST of a module.

    Args:
        module_name: The name of the module.

    Returns:
        The AST of the module.
    """
    # Import the module
    module = importlib.import_module(module_name)

    # Get the source code of the module
    module_file = inspect.getsourcefile(module)
    with open(module_file, "r") as file:
        module_source = file.read()

    # Parse the module source code into an AST
    module_ast = ast.parse(module_source)
    return module_ast

def write_module_source(module_name: str, source_code: str):
    """Write the source code of a module.

    Args:
        module_name: The name of the module.
        source_code: The source code of the module.
    """
    # Import the module.
    module = importlib.import_module(module_name)

    # Get the source code of the module.
    module_file = inspect.getsourcefile(module)

    # Format the source code using Black.
    source_code = black.format_str(source_code, mode=black.FileMode())

    # Write the new source code back to the module file.
    with open(module_file, "w") as file:
        file.write(source_code)


def get_function_source(module_name: str, function_name: str) -> str:
    """Get the source code of a function in a module.
    
    Args:
        module_name: The name of the module.
        function_name: The name of the function.
    
    Returns:
        The source code of the function.
    """
    # Import the module.
    module = importlib.import_module(module_name)

    # Get the function object using its name
    func_obj = getattr(module, function_name)

    # Get the source code of the function
    return inspect.getsource(func_obj)


def update_function(module_name: str, function_name: str, new_code: str):
    """Update the source code of a function in a module.
    
    Args:
        module_name: The name of the module.
        function_name: The name of the function.
        new_code: The new source code of the function.

    """
    module_ast = get_module_ast(module_name)

    # Parse the new function code and check for multiple function definitions
    new_code_ast = ast.parse(new_code)
    function_defs = [
        node for node in new_code_ast.body if isinstance(node, ast.FunctionDef)
    ]
    if len(function_defs) > 1:
        raise ValueError(
            "New function source must not contain more than one function definition."
        )

    # Define a class to visit and update the target function node
    class FunctionUpdater(ast.NodeTransformer):
        def visit_FunctionDef(self, node):
            if node.name == function_name:
                # Parse the new function code into an AST and replace the body
                node.body = function_defs[0].body
            return self.generic_visit(node)

    # Transform the AST
    updated_ast = FunctionUpdater().visit(module_ast)

    # Generate the new source code from the updated AST
    new_source_code = ast.unparse(updated_ast)

    # Validate the new source code. This is necessary to ensure the new code is valid Python syntax.
    env = {}
    exec(new_source_code, env, env)
    # Evaluate the new source code to ensure it is valid Python syntax.
    func = env.get(function_name)
    func()

    # Write the new source code back to the module file
    write_module_source(module_name, new_source_code)


def create_function(module_name: str, function_name: str, new_code: str):
    """Create a new component function in the module. Do this to create a new component before using it.

    Args:
        module_name: The name of the module to add the function to.
        function_name: The name of the new function.
        new_code: The source code of the new function.
    """
    module_ast = get_module_ast(module_name)

    # Create a new function definition node from the new code
    new_function_ast = ast.parse(new_code).body[0]

    # Ensure the parsed new code is a function definition
    if not isinstance(new_function_ast, ast.FunctionDef):
        raise ValueError("New code does not define a function.")

    # Update the function name if necessary
    new_function_ast.name = function_name

    # Add the new function to the module's AST
    module_ast.body.append(new_function_ast)

    # Generate the new source code from the updated AST
    new_source_code = ast.unparse(module_ast)
    write_module_source(module_name, new_source_code)



"""Welcome to Reflex! This file create a counter app."""

import os
import reflex as rx
from reflex_ai.selection import ClickSelectionState
from reflex_ai.local_agent import get_agent, InternRequest, InternResponse, ToolRequestResponse

import httpx
from flexai.message import Message, ToolUseMessage
from anthropic.types import ToolUseBlock

backend = "localhost:8000"

class PlaygroundState(rx.State):
    """The app state."""

    dev_mode: bool = True
    processing: bool = False

    selected_id: str = ""
    code: str = ""
    prompt: str = ""

    async def process(self, prompt):
        # Generate new code.
        self.processing = True
        yield

        selection_state = await self.get_state(ClickSelectionState)
        selected_code = "\n".join(selection_state._selected_code)

        unconverted_messages = []

        request = InternRequest(
            prompt=prompt["prompt"],
            selected_code=selected_code,
            selected_module=selection_state.selected_module,
            selected_function=selection_state.selected_function,
        )

        resp = httpx.post(
            f"http://{backend}/api/intern",
            data=request.model_dump_json(),
            timeout=60,
        )
        print(resp)
        print(resp.json())
        resp_obj = InternResponse(**resp.json())
        messages = [Message(role=m.role, content=m.content) for m in resp_obj.messages]

        while True:
            tool_response_messages = []
            local_intern = get_agent()
            for message in messages:
                try:
                    tool_use_message = ToolUseMessage.from_tool_use_block(
                        ToolUseBlock.parse_raw(message.content),
                    )
                except ValueError:
                    unconverted_messages.append(message)
                    continue
                tool_response_messages.append(
                    await local_intern.invoke_tool(tool_use_message)
                )

            if not tool_response_messages:
                break

            tool_response_request = ToolRequestResponse(
                request_id=resp_obj.request_id,
                messages=tool_response_messages,
            )
            print("REQUEST TO API: ", tool_response_request)

            resp2 = httpx.post(
                f"http://{backend}/api/intern/tool_response",
                data=tool_response_request.model_dump_json(),
                timeout=60,
            )
            resp2.raise_for_status()
            print("RESPONSE FROM API: ", resp2.content)
            messages = [Message(role=m["role"], content=m["content"]) for m in resp2.json()]

        self.processing = False

    async def handle_upload(self, files: list[rx.UploadFile]):
        """Handle the upload of file(s).

        Args:
            files: The uploaded files.
        """
        print("uploading")
        import base64
        for file in files:
            upload_data = await file.read()
            # Convert to base64.
            image_data = base64.b64encode(upload_data).decode()
            self.image = f"data:image/png;base64,{image_data}"

    async def get_source(self, data):
        """Get the source of the data."""
        if not open:
            return
        import importlib.util

        self.selected_id = data
        yield
        filename, function_name = data.split(":")
        module_name = os.path.splitext(os.path.basename(filename))[0]

        print(module_name)

        spec = importlib.util.spec_from_file_location(module_name, filename)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        function = getattr(module, function_name)
        if inspect.isfunction(function) or inspect.ismethod(function):
            self.code = inspect.getsource(function)
        else:
            print(f"{function_name} is not a function or method in {filename}")


ClientState = rx._x.client_state("code", rx.Var.create("code", _var_is_string=True))

import inspect
import os

def component(fn):
    """A decorator that adds an on_click handler to the component."""


def get_current_function_name():
    filename = __file__
    function_name = inspect.currentframe().f_back.f_code.co_name
    return f"{filename}:{function_name}"


def toolbar():
    return rx.hstack(
        rx.cond(
            PlaygroundState.processing,
            rx.spinner(size="3", color="white"),
        ),
        rx.upload.root(
            rx.button("Image"),
            id="image",
            on_drop=PlaygroundState.handle_upload(rx.upload_files(upload_id="image")),
        ),
        rx.form(
            rx.input(name="prompt", disabled=PlaygroundState.processing),
            on_submit=PlaygroundState.process,
            reset_on_submit=True,
        ),
    )

def playground(page):
    """The main view."""
    return rx.vstack(
        page(),
        rx.cond(
            PlaygroundState.dev_mode,
            rx.box(
                toolbar(),
                width="100%",
                bottom="0",
            ),
        ),
        rx.code_block(ClickSelectionState.code),
        min_height="100vh",
    )
